import os.path
import sys
import json
import time
import requests
from collections import namedtuple
from datetime import datetime

from rich.progress import track

from bakasur.exceptions import ThuisbezorgdApiError, ThuisbezorgdAuthError
from bakasur.utils import get_headers, get_tokens, update_token_file
from bakasur.database import get_latest_datetime_from_orders, BakasurDB
from bakasur.constants import (THUISBEZORGD_LOGIN_URL,
                               THUISBEZORGD_USER_URL,
                               THUISBEZORGD_ORDER_URL,
                               THUISBEZORGD_REFRESH_TOKEN,
                               THUISBEZORGD_TOKEN_FILE,
                               THUISBEZORGD_ORDER_DETAILS_URL,
                               HEADERS,
                               console)


class Login:
    def __init__(self, username: str, password: str) -> None:
        self.username = username
        self.password = password
        self.access_token = None
        self.refresh_token = None
        self.headers = HEADERS

    def save_token_file(self):
        data = {
            'authToken': self.access_token,
            'refreshToken': self.refresh_token,
        }
        with open(THUISBEZORGD_TOKEN_FILE, 'w') as token_file:
            json.dump(data, token_file)

    def attempt_login(self, vcode: str = None):
        status = 'LOGIN FAILED'

        if vcode is None:
            json_data = {
                'username': self.username,
                'password': self.password,
            }
        else:
            json_data = {
                'username': self.username,
                'password': self.password,
                'twoFactorAuthenticationCode': vcode,
            }

        try:
            resp = requests.post(THUISBEZORGD_LOGIN_URL, headers=self.headers, json=json_data)
            if resp.status_code == 201:
                console.print(
                    ":stopwatch: [yellow] Login initiated, you will now receive a verification code via mail. Your code expires in 30min")
                status = 'LOGIN PENDING'
            elif resp.status_code == 200:
                console.print(":party_popper: [green]Login successful!")
                data = resp.json()
                self.access_token = data.get('authToken', None)
                self.refresh_token = data.get('refreshToken', None)
                self.save_token_file()
                status = 'DONE'
        except ThuisbezorgdAuthError:
            sys.exit("Login attempt failed. Please check your credentials.")
        return status


def is_authenticated(func):
    def wrapper(*args, **kwargs):
        max_retries = 0
        while max_retries < 3:
            try:
                # print("Authenticating...")
                headers = get_headers()
                r = requests.get(THUISBEZORGD_USER_URL, headers=headers)
                if r.status_code == 200:
                    response = func(*args, **kwargs)
                    return response
                else:
                    print("Refreshing tokens")
                    r = requests.post(THUISBEZORGD_REFRESH_TOKEN, headers=HEADERS, json=get_tokens())
                    resp = r.json()
                    attempt = 1
                    while 'error' in resp.keys() and attempt < 3:
                        r = requests.post(THUISBEZORGD_REFRESH_TOKEN, headers=HEADERS, json=get_tokens())
                        time.sleep(2)
                        if r.status_code == 200:
                            resp = r.json()
                            break
                        attempt += 1
                    new_tokens = resp
                    update_token_file(new_tokens)
                    max_retries += 1
            except Exception as e:
                raise ThuisbezorgdAuthError(e)

    return wrapper


@is_authenticated
def fetch_orders(page=1, limit=100):
    # Fetch order data
    print("Fetching orders...")
    params = {
        'page': page,
        'limit': limit,
    }
    req = requests.get(THUISBEZORGD_ORDER_URL, headers=get_headers(), params=params)
    order_data = req.json()
    return order_data


@is_authenticated
def fetch_order_details(order_id):
    params = {
        'orderId': order_id
    }
    req = requests.get(THUISBEZORGD_ORDER_DETAILS_URL, headers=get_headers(), params=params)
    order_detail_data = req.json()
    return order_detail_data


def process_and_store_orders(db, orders_data, store_recent=False):
    orders = []
    Orders = namedtuple(
        'Orders', ['order_id', 'order_val', 'order_datetime', 'restaurant_name'])

    for key, data in orders_data.items():
        order_data = data.get('orders')
        for order in order_data:
            order_id = order.get('id')
            divisor = order['currency']['denominator']
            order_val = order.get('value') / divisor
            order_datetime = datetime.fromtimestamp(order.get('date') / 1000)
            restaurant_name = order['restaurant']['name']
            orders.append(Orders(order_id=order_id,
                                 order_val=order_val,
                                 order_datetime=order_datetime,
                                 restaurant_name=restaurant_name))

    if store_recent:
        # get the recent timestamp from the order table
        # compare the max
        db_datetime = get_latest_datetime_from_orders(db)
        orders = [order for order in orders if
                  order.order_datetime > datetime.strptime(db_datetime, "%Y-%m-%d %H:%M:%S")]
        if len(orders) == 0:
            print("There are no recent orders in your history since the last time")
            return
        else:
            db.insert_orders(orders)
            print("Storing your recent orders")
    else:
        db.insert_orders(orders)

    order_ids = [ord.order_id for ord in orders]
    item_dict = {}
    for idx, ord_id in track(enumerate(order_ids), description="Fetching order details..."):
        data = fetch_order_details(ord_id)
        item_dict[idx] = data
        time.sleep(2)

    order_details = []
    OrderDetails = namedtuple('OrderDetails', ['order_id', 'restaurant_name', 'order_name'])
    print("Starting to process the order details")
    for key, val in item_dict.items():
        orderID = val.get('order').get('id')
        restaurant_name = val.get('restaurant').get('name')
        cart_items = val.get('order').get('cartItems')
        for item in cart_items:
            item_name = item.get('name')
            order_details.append(OrderDetails(order_id=orderID,
                                              restaurant_name=restaurant_name,
                                              order_name=item_name))
    db.insert_order_details(order_details)


def store_orders(db, store_recent):
    try:
        data = {}
        orders = fetch_orders()
        total_pages = orders.get('pagination').get('totalPages')
        data[0] = orders

        if total_pages > 1:
            for page in range(2, total_pages + 1):
                orders = fetch_orders(page=page)
                time.sleep(2)
                data[page - 1] = orders

        process_and_store_orders(db, data, store_recent)
    except Exception as e:
        raise ThuisbezorgdApiError(e)



class ThuisbezorgdApiError(Exception):
    pass


class ThuisbezorgdAuthError(Exception):
    pass


class ThuisbezorgdTokenFileError(Exception):
    pass


class ThuisbezorgdCliDBError(Exception):
    pass



import os
import json
from bakasur.constants import (THUISBEZORGD_TOKEN_FILE,
                               DB_FILEPATH,
                               HEADERS)


def token_file_exists():
    """
    Checks whether the token file is present or not
    """
    return os.path.exists(os.path.join(THUISBEZORGD_TOKEN_FILE))


def db_file_exists():
    """
    Checks whether the DB file is present or not
    """
    return os.path.exists(os.path.join(DB_FILEPATH))


def get_headers():
    """
    Adds the authorization token to the headers
    """
    with open(THUISBEZORGD_TOKEN_FILE, 'r') as token_file:
        f = json.load(token_file)
        auth_token = f['authToken']
    headers = HEADERS.copy()
    headers['authorization'] = 'Bearer ' + auth_token
    return headers


def get_tokens():
    """
    Return json data for refresh tokens request
    """
    with open(THUISBEZORGD_TOKEN_FILE, 'r') as token_file:
        tokens = json.load(token_file)
    return {
        'accessToken': tokens.get('authToken'),
        'refreshToken': tokens.get('refreshToken'),
    }


def update_token_file(data):
    """
    Write the newly received tokens to the token for future use
    """
    # print("Writing file")
    with open(THUISBEZORGD_TOKEN_FILE, 'w') as outfile:
        json.dump(data, outfile)



import typer

from bakasur.constants import console
from bakasur.helpers import Login, store_orders
from bakasur.utils import token_file_exists, db_file_exists
from bakasur.database import BakasurDB
from bakasur.dashboard import display

app = typer.Typer()


@app.callback(invoke_without_command=True)
def run():
    console.rule("[bold red] Welcome to Bakasur!")
    console.print(
        'Bakasur is your friendly demon :ogre: who helps you track your Thuisbezorgd order history and visualise it.',
        style="bold white", justify="left", crop=True)
    console.rule("[bold red] ***")

    if not token_file_exists():
        user = typer.prompt("Please enter your email id", type=str)
        password = typer.prompt("Please enter your password", type=str, hide_input=True)

        loginObj = Login(username=user, password=password)
        result = loginObj.attempt_login()
        if result == 'LOGIN PENDING':
            # console.log(result)
            vcode = typer.prompt("Please enter the verification code", type=str)
            loginObj.attempt_login(vcode=vcode)
        else:
            console.print(":warning: [red]Login attempt failed. Please check your credentials.")
            return

    if db_file_exists():
        db = BakasurDB()
        db.init_db()
        store_orders(db, store_recent=True)
    else:
        db = BakasurDB()
        db.init_db()
        db.create_db()
        store_orders(db, store_recent=False)

    orders_df, order_details_df = db.export_to_df()
    display(orders_df, order_details_df)


@app.command()
def reset():
    """
    This command deletes the DB, token files to start a new session
    """
    pass


if __name__ == "__main__":
    app()



import os
from rich.console import Console

THUISBEZORGD_URL = "https://cw-api.takeaway.com"
THUISBEZORGD_USER_URL = THUISBEZORGD_URL + "/api/v31/user"
THUISBEZORGD_LOGIN_URL = THUISBEZORGD_URL + "/api/v31/user/login" #POST 
THUISBEZORGD_ORDER_URL = THUISBEZORGD_URL + "/api/v31/user/orders" #GET
THUISBEZORGD_ORDER_DETAILS_URL = THUISBEZORGD_URL + "/api/v31/order/details"
THUISBEZORGD_REFRESH_TOKEN = THUISBEZORGD_URL + "/api/v31/user/jwt_refresh" #POST

DB_FILEPATH = os.path.join(str(os.getcwd()), 'thuisbezorgd.db')

# Credentials file
THUISBEZORGD_TOKEN_FILE = os.path.join(str(os.getcwd()), 'thuisbezorgd_token.json')

REPORT_PATH = os.path.join(str(os.getcwd()), 'report.html')

console = Console()

HEADERS = {
            'authority': 'cw-api.takeaway.com',
            'accept': 'application/json, text/plain, */*',
            'accept-language': 'en',
            'content-type': 'application/json;charset=UTF-8',
            'origin': 'https://www.thuisbezorgd.nl',
            'referer': 'https://www.thuisbezorgd.nl/',
            'user-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.0.0 Safari/537.36',
            'x-country-code': 'nl',
            'x-language-code': 'en',
            'x-requested-with': 'XMLHttpRequest',
        }



import sqlite3
import pandas as pd

from bakasur.constants import DB_FILEPATH
from bakasur.exceptions import ThuisbezorgdCliDBError

create_order_table = """
CREATE TABLE IF NOT EXISTS orders(
id INTEGER PRIMARY KEY AUTOINCREMENT,
order_id INT UNIQUE,
order_val FLOAT,
order_datetime DATETIME,
restaurant_name VARCHAR)
"""

create_order_details_table = """
CREATE TABLE IF NOT EXISTS items(
id INTEGER PRIMARY KEY AUTOINCREMENT,
order_id INT,
restaurant_name VARCHAR,
order_name VARCHAR,
FOREIGN KEY(order_id) REFERENCES orders(order_id))
"""

latest_datetime_from_orders = """
SELECT order_datetime FROM orders ORDER BY order_datetime desc LIMIT 1
"""

insert_orders_query = """
INSERT INTO orders(order_id, order_val, order_datetime, restaurant_name) VALUES (?,?,?,?)
"""

insert_order_details_query = """
INSERT INTO items(order_id, restaurant_name, order_name) VALUES (?,?,?)
"""


def get_latest_datetime_from_orders(db):
    return db.fetch_result(query=latest_datetime_from_orders)[0][0]


class BakasurDB:
    def init_db(self):
        try:
            print("Connecting to {}".format(DB_FILEPATH))
            self.conn = sqlite3.connect(DB_FILEPATH)
        except Exception:
            raise ThuisbezorgdCliDBError("Unable to connect to database")

    def create_db(self):
        cur = self.conn.cursor()
        cur.execute(create_order_table)
        cur.execute(create_order_details_table)
        self.conn.commit()

    def insert_orders(self, orders):
        cur = self.conn.cursor()
        try:
            cur.executemany(insert_orders_query, orders)
            self.conn.commit()
        except sqlite3.Error as e:
            if "UNIQUE" in "{}".format(e):
                pass
            else:
                raise ThuisbezorgdCliDBError("Error while inserting orders: {}".format(e))

    def insert_order_details(self, order_details):
        cur = self.conn.cursor()
        try:
            cur.executemany(insert_order_details_query, order_details)
            self.conn.commit()
        except sqlite3.Error as e:
            if "UNIQUE" in "{}".format(e):
                pass
            else:
                raise ThuisbezorgdCliDBError("Error while inserting order details: {}".format(e))

    def fetch_result(self, query):
        cur = self.conn.cursor()
        try:
            response = cur.execute(query).fetchall()
            self.conn.commit()
        except sqlite3.Error as e:
            raise ThuisbezorgdCliDBError("Error while fetching items: {}".format(e))
        except Exception as e:
            raise ThuisbezorgdCliDBError("Error while executing query: {}".format(e))
        return response

    def export_to_df(self):
        # cur = self.conn.cursor()
        try:
            orders = self.conn.execute("SELECT * FROM orders")
            order_details = self.conn.execute("SELECT * FROM items")

            orders_cols = [column[0] for column in orders.description]
            order_detail_cols = [column[0] for column in order_details.description]

            orders_df = pd.DataFrame.from_records(data=orders.fetchall(), columns=orders_cols)
            order_details_df = pd.DataFrame.from_records(data=order_details.fetchall(), columns=order_detail_cols)
        except Exception as e:
            raise ThuisbezorgdCliDBError("Error while executing query: {}".format(e))

        return orders_df, order_details_df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bakasur']

package_data = \
{'': ['*']}

install_requires = \
['datapane>=0.14.0,<0.15.0',
 'requests>=2.28.1,<3.0.0',
 'rich>=12.4.4,<13.0.0',
 'typer[all]>=0.4.2,<0.5.0']

entry_points = \
{'console_scripts': ['bakasur = bakasur.main:app']}

setup_kwargs = {
    'name': 'bakasur',
    'version': '1.3.0',
    'description': 'Bakasur is your friendly demon that helps you analyse your Thuisbezorgd order history and visualise patterns.',
    'long_description': '# Bakasur\n[![PyPI version](https://badge.fury.io/py/bakasur.svg)](https://badge.fury.io/py/bakasur)\n***\n[Bakasur](https://github.com/d-kold/bakasur) (_/buh-KAA-soor/_) is your friendly demon who helps you analyse your Thuisbezorgd order history and visualise it. \n\n![Dashboard](img/dashboard.png)\n\n![Terminal](img/terminal.png)\n\n\n### Disclaimer\n***\nThis tool requires you to log in using your Thuisbezorgd credentials. Upon correct input, you will recieve \na email with a verification code from Thuisbezorgd. Once the details are added, `bakasur` will create a `thuisbezorgd_token.json`\nfile in your current working directory which will store the `authToken` and `refreshToken` received from the login request. \nThese tokens will help in authenticating the Thuisbezorgd API in the subsequent runs of this tool. \nIf you are concerned about the security of the tokens they are on your local filesystem and you can also review\nthe code to check that there are no evil intentions here. **Your username and password are not stored.**\n\n\n### How to use this tool\n***\nThis tool uses a number of packages such as `typer`, `rich`, `datapane`. To avoid any conflicts with your current installed\nsite-packages, it is preferable to create a virtual environment and then install this tool.\n\n- Create a virtual environment `virtualenv venv` and activate it `source venv/bin/activate`\n- Install Bakasur using `pip install bakasur`\n- Once it is installed you can now type in `bakasur` in your terminal and get started\n\n### Tool specifics\n***\nWhen you first log in to Thuisbezorgd using `bakasur` it creates a sqlite database in your current working directory `thuisbezrgd.db`. \nThe database stores all your orders and their details in tables `orders` and `items` respectively. During subsequent runs\nof the tool, your most recent orders are inserted into the database \n\nRequires:\n- Python v3.7.1+ \n- Thuisbezorgd account (duh!)\n\n### Contribution\n***\nGive this tool a try and if you find any bugs or issues with it then feel free to open an issue or start a discussion.\nFeature requests, Bug fixes are most welcome.\n\n### License\n***\n[MIT License](https://github.com/d-kold/bakasur/blob/0d2317c116180b2e33d14e833c25352ff5a8e032/LICENSE.md)\n\n### Important\n***\nThis tool is intended for personal use. A fun hobby project for demo. I am not responsible if you tweak the code and violate \nany Thuisbezorgd Terms and Conditions. \n\n[d-kold](https://github.com/d-kold)',
    'author': 'Devendra Kulkarni',
    'author_email': 'kulkarnidevendra21@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<3.11.0',
}


setup(**setup_kwargs)



import pandas as pd
import datapane as dp
import altair as alt
from datetime import datetime

from bakasur.constants import REPORT_PATH

pd.options.mode.chained_assignment = None


def display(order_data, order_details_data):
    order_data['order_datetime'] = pd.to_datetime(order_data.order_datetime)
    order_data['weekday'] = order_data.order_datetime.dt.day_name()
    days_since_last_order = pd.to_datetime(datetime.today()) - order_data.order_datetime.max()

    weekday_dist_df = order_data['weekday'].value_counts().rename_axis('weekday').reset_index(name='counts')
    weekday_dist_df['weekday'] = pd.Categorical(weekday_dist_df['weekday'],
                                                categories=['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday',
                                                            'Saturday', 'Sunday'],
                                                ordered=True)
    weekday_dist_df = weekday_dist_df.sort_values('weekday').reset_index(drop=True)

    spends_per_month_df = order_data[["order_datetime", "order_val"]]
    spends_per_month_df["order_count"] = 1
    spends_per_month_df = spends_per_month_df.resample("M", on="order_datetime").sum().reset_index()
    spends_per_month_df = spends_per_month_df.assign(month_year=lambda x: x['order_datetime'].dt.strftime("%b-%y"))

    top_restaurants_df = order_details_data['restaurant_name'].value_counts().rename_axis('restaurant').reset_index(
        name='num_ordered').query("num_ordered > 3")

    most_ordered_item_df = order_details_data["order_name"].value_counts().rename_axis('item').reset_index(
        name='num_ordered').query("num_ordered >= 2")

    most_ordered_item_chart = alt.Chart(most_ordered_item_df).mark_bar().encode(
        x=alt.X('num_ordered', sort=None, title="Number of orders"),
        y=alt.Y('item', sort=None, title="Item"),
        color=alt.Color('num_ordered',
                        scale=alt.Scale(scheme='teals'))
    ).properties(title="My most ordered dishes")

    top_restaurants_chart = alt.Chart(top_restaurants_df).mark_bar().encode(
        x=alt.X('num_ordered', sort=None, title="Number of orders"),
        y=alt.Y('restaurant', sort=None, title="Restaurant"),
        color=alt.Color('num_ordered',
                        scale=alt.Scale(scheme='teals'))
    ).properties(title="Restaurants I frequently order from")

    base = alt.Chart(spends_per_month_df).mark_bar().encode(
        x={
            "field": "order_val",
            "type": "quantitative",
            "title": "Spends per month (in €)"
        },
        y={
            "field": "month_year",
            "type": "ordinal",
            "sort": None,
            "title": "Month-Year"
        }
    )

    spends_chart = base.mark_bar().encode(
        color=alt.Color('order_val',
                        scale=alt.Scale(scheme='lightmulti'))
    ).properties(title="Money spent and orders placed each month")

    text = base.mark_text(
        align='left',
        baseline='middle',
        dx=3
    ).encode(
        text='order_count'
    )

    alt_chart = alt.Chart(weekday_dist_df).mark_bar().encode(
        x={"field": "weekday",
           "type": "ordinal",
           "sort": ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday',
                    'Saturday', 'Sunday'],
           "title": "Weekday"},
        y={"field": "counts",
           "type": "quantitative",
           "title": "Orders per day"},
        color=alt.Color('counts',
                        scale=alt.Scale(scheme='blues'))).properties(title="When do I order most frequently?")
    dp.Report(
        dp.Group(
            dp.BigNumber(
                heading="Days since my last order", value=days_since_last_order.days),
            dp.BigNumber(heading="Total number of orders", value=len(order_data)),
            columns=2
        ),
        dp.Group(
            dp.Plot(alt_chart),
            dp.Plot((spends_chart + text).properties(height=400)),
            columns=2
        ),
        dp.Group(
            dp.Plot(top_restaurants_chart),
            dp.Plot(most_ordered_item_chart),
            columns=2
        )
    ).save(open=True, path=REPORT_PATH)
